# -------------------------------------------------------------------
# Full moment tensor
#
# refs
# Tape, W., & Tape, C. (2015). A uniform parametrization of moment tensors.
# Geophysical Journal International, 202(3), 2074–2081.
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

from seisgen.MTTools.DLune import DCreate_Triple
from seisgen.MTTools.DOrientations import DRotationMatrix, DRotz
import numpy as np

gRMax = DRotz(np.deg2rad(-45))
ginvRMax = np.linalg.inv(gRMax)

def DMomentTensor(strike, dip, rake, colatitude, longitude):
    '''
    Calculate the FULL moment tensor for the given strike, dip, rake and the colatitude, longitude in Lune.

    :param strike:  The strike in radians, k in Carl's paper.  ranging [0, 2pi]
    :param dip:     The dip in radians, thetha in Carl's paper. ranging [0, pi/2]
    :param rake:    The slip (rake) in radians, sigma in Carl's paper. ranging [-pi/2, pi/2]
    :param colatitude: The colatitude in the Lune in radians (Beta), [0, PI]
    :param longitude:  The longitude in the Lune in radians (Gamma), [-PI/6, PI/6]
    :return: The full moment tensor.
    '''

    triple = DCreate_Triple(colatitude, longitude)
    triple_matrix = np.zeros((3, 3))
    triple_matrix[0, 0] = triple[0]
    triple_matrix[1, 1] = triple[1]
    triple_matrix[2, 2] = triple[2]

    U = DRotationMatrix(strike, dip, rake)
    return np.dot(np.dot(U, triple_matrix), np.linalg.inv(U))


def DMT_enz(strike, dip, rake, colatitude, longitude):
    ''' return the moment tensor in e-n-z coord.'''
    mt_nwz = DMomentTensor(np.deg2rad(strike), np.deg2rad(dip), np.deg2rad(rake),
                           np.deg2rad(colatitude), np.deg2rad(longitude))
    mt_enz = np.ones(6)
    mt_enz[0] = mt_nwz[1][1]
    mt_enz[1] = mt_nwz[0][0]
    mt_enz[2] = mt_nwz[2][2]

    mt_enz[3] = -1.0 * mt_nwz[0][1]
    mt_enz[4] = -1.0 * mt_nwz[1][2]
    mt_enz[5] = mt_nwz[0][2]

    return mt_enz * np.sqrt(2)



# -------------------------------------------------------------------
# ibool reader.
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

from seisgen.util_SPECFEM3D import NGLLX, NGLLY, NGLLZ, CONSTANT_INDEX_27_GLL
from scipy.io import FortranFile
import numpy as np


def read_ibool_by_scipy(ibool_file, NSPEC):
    '''
        Read the ibool file in the folder */model3D/
    '''

    f = FortranFile(ibool_file, 'r')
    ibool = f.read_reals(dtype='int32')
    f.close()
    ibool = np.reshape(ibool, (NSPEC, NGLLX * NGLLY * NGLLZ))

    # The index in *.ibool files starts from 1.
    ibool = ibool - 1
    return ibool


def DEnquire_Element(ibool_file, index_element, NSPEC):
    ''' Read the index of the 27 GLL points where the SGT been stored in the selected element.'''

    ibool = read_ibool_by_scipy(ibool_file, NSPEC)
    if ibool.__len__() <= index_element:
        return np.zeros(27)

    else:
        NGLLX_N3 = 3
        NGLLY_N3 = 3
        NGLLZ_N3 = 3

        # The global index in slice of selected GLL points.
        gll_array = ibool[index_element][CONSTANT_INDEX_27_GLL]

        # sort the index.
        gll_points = []
        gll_array = np.reshape(gll_array, [NGLLZ_N3, NGLLY_N3, NGLLX_N3])
        for i in range(NGLLX_N3):
            for j in range(NGLLY_N3):
                for k in range(NGLLZ_N3):
                    gll_points.append(gll_array[k, j, i])
        gll_points = np.asarray(gll_points)

        return gll_points



# -------------------------------------------------------------------
# Tools to manage the pre-computed grids in the 3D background model.
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import h5py
import numpy as np

POINT_KEYS = ["latitude",
             "longitude",
             "z",
             "depth",
             "utm_x",
             "utm_y",
             "utm_z",
             "slice_index",
             "element_index",
             "xi",
             "eta",
             "gamma"]


class DPointCloud():
    ''' The pre-computed information (see POINT_KEYS) of user-selected points in the 3D background model. '''

    def __init__(self, file_path):
        '''
        :param file_path:  The HDF5 file storing the pre-computed information (see POINT_KEYS)
                           of user-selected points in the 3D background model.
        '''

        self.b_pointcloud_initial = False
        if file_path is None:
            return

        try:
            with h5py.File(file_path, 'r') as f:
                self.mesh_lat           = f[POINT_KEYS[0]][:]
                self.mesh_long          = f[POINT_KEYS[1]][:]
                self.mesh_z             = f[POINT_KEYS[2]][:]    # in meter
                self.mesh_depth         = f[POINT_KEYS[3]][:]    # in meter.
                self.mesh_utm_x         = f[POINT_KEYS[4]][:]
                self.mesh_utm_y         = f[POINT_KEYS[5]][:]
                self.mesh_utm_z         = f[POINT_KEYS[6]][:]
                self.mesh_slice_index   = f[POINT_KEYS[7]][:]
                self.mesh_element_index = f[POINT_KEYS[8]][:]
                self.mesh_xi            = f[POINT_KEYS[9]][:]
                self.mesh_eta           = f[POINT_KEYS[10]][:]
                self.mesh_gamma         = f[POINT_KEYS[11]][:]

            self.n_grid = len(self.mesh_lat)
            self.b_pointcloud_initial = True
        except:
            print("!!! Point cloud not found")
            raise Exception


    def _check(self):
        if self.b_pointcloud_initial is False:
            print("!!! Point cloud not initialized")
            raise Exception


    def find(self, x, y, z, n=1, mode='LATLONGZ', b_depth=True):
        '''
        Determine the closest N points to (x, y, z) in the point cloud and return the information.
         :param x:       Either the latitude or UTMX.
         :param y:       Either the longitude or UTMY.
         :param z:       The depth or elevation in meter. eg: depth: 2000, elevation: -1200
         :param n:       The number of point enquired.
         :return:        The information (see POINT_KEYS) of the closest N points.
        '''

        self._check()

        n = int(n)
        n_point = len(self.mesh_lat)
        distance_arr = np.zeros(n_point)

        if str(mode).upper() == str('LATLONGZ'):
            if b_depth:
                # use depth.
                for i in range(n_point):
                    _dist_H = 111.0 * 1000.0 * np.sqrt(np.square(self.mesh_lat[i] - x) + np.square(self.mesh_long[i]-y))
                    distance_arr[i] = np.sqrt(np.power(_dist_H, 2) + np.power(self.mesh_depth[i] - z, 2))
            else:
                # use elevation.
                for i in range(n_point):
                    _dist_H = 111.0 * 1000.0 * np.sqrt(np.square(self.mesh_lat[i] - x) + np.square(self.mesh_long[i]-y))
                    distance_arr[i] = np.sqrt(np.power(_dist_H, 2) + np.power(self.mesh_z[i] - z, 2))

        elif str(mode).upper() == str('UTM'):
            if b_depth:
                # use depth
                for i in range(n_point):
                    distance_arr[i] = np.sqrt(np.power(self.mesh_utm_x[i] - x, 2)
                                              + np.power(self.mesh_utm_y[i] - y, 2)
                                              + np.power(self.mesh_depth[i] - z, 2))
            else:
                # use elevation.
                for i in range(n_point):
                    distance_arr[i] = np.sqrt(np.power(self.mesh_utm_x[i] - x, 2)
                                              + np.power(self.mesh_utm_y[i] - y, 2)
                                              + np.power(self.mesh_z[i] - z, 2))
        else:
            print("!!! Undefined mode!")
            raise NotImplementedError

        if 1 == n:
            idx = np.argmin(distance_arr)
        else:
            idx = np.argpartition(distance_arr, n)[:n]

        return self.mesh_lat[idx], self.mesh_long[idx], self.mesh_z[idx], self.mesh_depth[idx],\
               self.mesh_utm_x[idx], self.mesh_utm_y[idx], self.mesh_utm_z[idx], \
               self.mesh_slice_index[idx], self.mesh_element_index[idx], \
               self.mesh_xi[idx], self.mesh_eta[idx], self.mesh_gamma[idx]




# -------------------------------------------------------------------
# Tools to acquire Greens Function (the displacement, DGF)
# from the 3D pre-stored receiver-side Greens function database
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------


import numpy as np
import h5py
import zlib
from seisgen.util_SPECFEM3D import SGT_ENCODING_LEVEL


DGF_KEYS = [
    'index',
    'start',
    'length',
    'offset',
    'scale',
]

DGF_ATTRS = [
    'ngll',
    'nstep',
    'nforce',
    'nparas',
    'dt',
    'nspec',
    'nGLL_global',
    'type',
]


def read_DGF_header(header_path):
    '''Acquire parameters in the header file. '''
    # load the information of about the GF database.
    with h5py.File(header_path, 'r') as f:
        dt      = f.attrs[DGF_ATTRS[4]]
        NSPEC   = f.attrs[DGF_ATTRS[5]]
        nGLL_global = f.attrs[DGF_ATTRS[6]]
    return dt, NSPEC, nGLL_global


def DEnquire_DGF(data_path, header_path, g_indx_GLL_points):
    '''
    * Enquire the GF from the database (*.bin files).

    :param data_path:           The path to the data file (.bin)
    :param header_path:         The path to the header file (.hdf5)
    :param g_indx_GLL_points:   The global index of the GLL points acquired in the slice.
    :return:                    The list of GF array acquired.
    '''

    with h5py.File(header_path, 'r') as f:
        names_GLL_arr       = f[DGF_KEYS[0]][:]
        data_start_array    = f[DGF_KEYS[1]][:]
        data_length_array   = f[DGF_KEYS[2]][:]
        data_offset_array   = f[DGF_KEYS[3]][:]
        data_scale_array    = f[DGF_KEYS[4]][:]

        n_gll   = f.attrs[DGF_ATTRS[0]]
        n_step  = f.attrs[DGF_ATTRS[1]]
        n_dim   = f.attrs[DGF_ATTRS[2]]
        n_paras = f.attrs[DGF_ATTRS[3]]
        dt      = f.attrs[DGF_ATTRS[4]]
        NSPEC   = f.attrs[DGF_ATTRS[5]]
        nGLL_global = f.attrs[DGF_ATTRS[6]]

    dgf_arr_list = []
    with open(data_path, "rb") as fr:
        for gll in g_indx_GLL_points:
            idx_gll = (np.where(names_GLL_arr == gll))[0][0]

            offset_min = data_offset_array[idx_gll]
            normal_factor = data_scale_array[idx_gll]
            sgt_begin_bytes = data_start_array[idx_gll]
            sgt_length_bytes = data_length_array[idx_gll]

            # extract the compressed data.
            fr.seek(sgt_begin_bytes)
            data = fr.read(sgt_length_bytes)

            # uncompress the data to bytes.
            data = zlib.decompress(data)

            # recover bytes to uint8
            data = np.frombuffer(data, dtype=np.uint8)

            # recover waveform
            data = data / (2 ** SGT_ENCODING_LEVEL - 1) * normal_factor + offset_min

            # data container
            unzip_sgt = np.zeros([n_step, n_dim, n_paras]).astype(np.float32)
            count = 0
            for j in range(n_dim):
                for k in range(n_paras):
                    unzip_sgt[:, k, j] = data[count * n_step:(count + 1) * n_step]
                    count += 1

            dgf_arr_list.append(unzip_sgt)

    return dgf_arr_list


# -------------------------------------------------------------------
# Widgets making life easier
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

# DO NOT CHANGE
# if you utilize the pyCAPLunar package to create the SGT database.
SGT_ENCODING_LEVEL = 8
NGLL_SPEC_COMPRESSION = 27

NGLLX = 5
NGLLY = NGLLX
NGLLZ = NGLLX
CONSTANT_INDEX_27_GLL = [0, 2, 4, 10, 12, 14, 20, 22, 24,
                         50, 52, 54, 60, 62, 64, 70, 72, 74,
                         100, 102, 104, 110, 112, 114, 120, 122, 124]


def get_proc_name(idx_processor):
    '''Return the processor name.'''
    return str('proc')+str(idx_processor).rjust(6, '0')



# -------------------------------------------------------------------
# Interpolation tools.
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import numpy as np

# Lagrange polynomial
def DCreate_anchors_xi_eta_gamma(ngll_xyz):
    '''
    * The xi, eta, and gamma array for the anchor points. (Constant)
    * the index of the location array: [Z, Y, X]

    :param ngll_xyz: The number of GLL point in each direction.
                     5 if all 125 GLL points are selected.
                     3 if the 27 designated GLL points are selected.
    :return:
            the xi, eta, and gamma array.
    '''

    if 3 == int(ngll_xyz):
        xi = np.array([-1.0, 0.0, 1.0])
    else:
        xi = np.array([-1.0, -1.0 * np.sqrt(3.0 / 7.0), 0.0, np.sqrt(3.0 / 7.0), 1.0])
    eta = xi.copy()
    gamma = xi.copy()
    return xi, eta, gamma


def DLagrange_any(xi, xi_gll):
    '''
    * compute the lagrange coefficient for the giving data.

    :param xi:      The data to be interpolated.
    :param x_gll:   The giving data. a set of point. np.array.
    :return: h_arr: The coefficient for the interpolation.
    '''

    nGll = len(xi_gll)
    h_arr = []
    for i in range(nGll):
        prod1 = 1.0
        prod2 = 1.0
        x0 = xi_gll[i]
        for j in range(nGll):
            if j == i:
                continue

            x = xi_gll[j]
            prod1 *= (xi - x)
            prod2 *= (x0 - x)

        h_arr.append(prod1/prod2)
    h_arr = np.asarray(h_arr)
    return h_arr


def DLagrange_any3D(xi, eta, gamma, xi_gll, eta_gll, gamma_gll):
    '''
    * compute the lagrange coefficient for the giving 3D data.

    :param xi: The relative x of the location to be interpolated.
    :param eta: The relative y of the location to be interpolated.
    :param gamma: The relative z of the location to be interpolated.
               The xi, eta, gamma are the relative location.
               Size: 1
               Data type: nd.float
    :param xi_gll: The x array of the knowing gll points.
                  Eg: Data size: NGLL=5.
                  Data type: np.array.
    :param eta_gll: The y array of the knowing points.
                    Eg: Data size: 5.
                    Data type: np.array.
    :param gamma_gll: The z array of the knowing points.
                      Eg: Data size: 5.
                      Data type: np.array.
    :return: coef
    Eg: Data size: 125, represents coefficient of the 125 gll points.
                  Data type: np.array.
    '''

    h_xi_arr = DLagrange_any(xi, xi_gll)
    h_eta_arr = DLagrange_any(eta, eta_gll)
    h_gamma_arr = DLagrange_any(gamma, gamma_gll)
    return h_xi_arr, h_eta_arr, h_gamma_arr


def DLagrange_interp_sgt(h_xi_arr, h_eta_arr, h_gamma_arr, sgt_arr_list, ngll_x, ngll_y, ngll_z):
    '''
    * Utilizing the coef_3d and the value array to interpolation.

    :param h_xi_arr:    The computed xi array for interpolation.
    :param h_eta_arr:   The computed eta array for interpolation.
    :param h_gamma_arr: The computed gamma array for interpolation.
    :param sgt_arr_list: List of the SGT data on the selected GLL points. Typically 125 GLL points.
                    sgt_arr size: 125 * [n_step, 3, 6], where
                    * The 125 represents the number of GLL points in one element.
                    * The greens_function order = ibool.
                    * The n_step represents the number of the data for the database.
                    * The 3 represents the number of force.
                    * The 6 represents the number of the element in SGT.
    :param ngll_x:    The number of GLL point in X direction.
                      data type: INT.
                      eg: 5 (NGLLX) if 125 GLL points are selected. 3 if 27 GLL points.
    :param ngll_y:    The number of GLL point in Y direction. As same as ngll_x.
    :param ngll_z:    The number of GLL point in Z direction. As same as ngll_x.
    :return:
            the interpolated SGT.
    '''

    n_step, n_dim, n_para = np.shape(sgt_arr_list[0])
    sgt_interp = np.zeros([n_step, n_dim, n_para])

    idx_sgt = 0
    for i in range(ngll_x):
        for j in range(ngll_y):
            for k in range(ngll_z):
                sgt_interp += h_xi_arr[i] * h_eta_arr[j] * h_gamma_arr[k] * sgt_arr_list[idx_sgt]
                idx_sgt += 1

    return sgt_interp



# -------------------------------------------------------------------
# Green's Function (Displacement, DGF) database Manager
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import os.path
from seisgen.DPointMgr import DPointCloud
from seisgen.util_SPECFEM3D import get_proc_name
from seisgen.util_SPECFEM3D.ibool_reader import DEnquire_Element
from seisgen.util_SPECFEM3D.xyz_reader import DEnquire_XYZ_GLLs_Element
from seisgen.greens_function.dgf_reader import DEnquire_DGF, read_DGF_header
from seisgen.math.interp_tools import DCreate_anchors_xi_eta_gamma, DLagrange_interp_sgt, DLagrange_any3D

from obspy.core.util.attribdict import AttribDict
from obspy.core import Stream, Trace
from obspy.clients.iris import Client

import numpy as np
import time


class DDGFMgr(DPointCloud):
    '''Greens Function (Displacement, DGF) database Manager'''

    def __init__(self, dgf_database_folder, model3D_folder, point_cloud_file, DLite=False):
        '''
        :param dgf_database_folder:     The directory to the greens function database.
        :param model3D_folder:          The directory to the 3D background model.
        :param point_cloud_file:        The hdf5 file storing the information of user-selected points in the 3D model.
        '''

        self.dgf_database_folder = dgf_database_folder
        self.model3D_folder = model3D_folder
        self.idx_element = -1
        self.dt = 0
        self.NSPEC = 0
        self.nGLL_global = 0
        # initial parameters of the GF database.
        if DLite:
            pass
        else:
            self.__initial_paras()
        super().__init__(point_cloud_file)

    def __initial_paras(self):
        '''Initializing the parameters.'''
        hdf5_files = []
        for (dirpath, dirnames, filenames) in os.walk(self.dgf_database_folder):
            for file in filenames:
                if str(file).lower().endswith('.hdf5') or str(file).lower().endswith('.h5'):
                    hdf5_files.append(os.path.join(dirpath, file))

        for file in hdf5_files:
            try:
                self.dt, self.NSPEC, self.nGLL_global = read_DGF_header(file)
                break
                
            except:
                continue

        if 0 == self.dt or 0 == self.NSPEC or 0 == self.nGLL_global:
            raise Exception


    def _initial_element_frame(self):
        ''' return the gll information (index, location) at one selected element. '''
        ibool_file = os.path.join(str(self.model3D_folder), str(self.proc_name) + str("_ibool.bin"))
        self.idx_glls = DEnquire_Element(ibool_file, self.idx_element, self.NSPEC)
        x_glls, y_glls, z_glls = DEnquire_XYZ_GLLs_Element(self.model3D_folder,
                                                           self.idx_processor,
                                                           self.idx_element,
                                                           self.NSPEC)
        self.xyz_glls = np.transpose(np.vstack([x_glls, y_glls, z_glls]))


    def _initial_DGFs_N_station(self):
        '''Return the GFs between origin and station. '''
        dir_string = os.path.join(str(self.dgf_database_folder),
                                  str(self.station.network),
                                  str(self.station.station),
                                  str(self.proc_name))
        data_path = dir_string + str("_dgf_data.bin")
        hder_path = dir_string + str("_dgf_header.hdf5")
        self.dgfs = DEnquire_DGF(data_path, hder_path, self.idx_glls)


    def interp_Lagrange(self, xi, eta, gamma):
        '''Using lagrange method to interpolate the GF data.'''
        ngll_x = 3
        ngll_y = 3
        ngll_z = 3

        xi_gll, eta_gll, gamma_gll = DCreate_anchors_xi_eta_gamma(ngll_xyz=3)
        h_xi_arr, h_eta_arr, h_gamma_arr = DLagrange_any3D(xi, eta, gamma, xi_gll, eta_gll, gamma_gll)

        self.dgf_interp = DLagrange_interp_sgt(h_xi_arr, h_eta_arr, h_gamma_arr, self.dgfs,
                                               ngll_x=ngll_x, ngll_y=ngll_y, ngll_z=ngll_z)

    def set_dt(self, dt):
        '''
        Set time interval.
        Function observed for SeisClient.
        '''
        self.dt = dt
        return self


    def get_dgf(self, station, origin, b_new_origin=True, b_verbose=False):
        '''
        Get the interpolated GF between the station-origin pair.
        Unit: m/N.m

        :param station: An instance of the obspy AttribDict class. For example:
                        station = AttribDict({ 'latitude': 34.0210,
                                                'longitude': -118.287,
                                                'network': 'CI',
                                                'station': 'USC',
                                                'location': '',
                                                'id': 'USC'})

        :param origin:  An instance of the obspy AttribDict class. For example:
                        origin = Origin({'time': '2019-07-04T18:39:44.0000Z',
                                              'latitude': 35.601333,
                                              'longitude': -117.597,
                                              'depth_in_m': 2810.0,
                                              'id': 'evt11056825'})

        :param b_new_origin: True, If acquiring GFs at multiple stations for a same origin to save time.
        '''

        if b_verbose:
            t0 = time.time()

        self.station = station

        if not self.b_pointcloud_initial:
            raise Exception

        try:
            z = origin.depth_in_m
            b_depth = True
        except:
            z = origin.elevation_in_m
            b_depth = False

        if b_new_origin:
            _, _, _, _, \
            _, _, _, \
            self.idx_processor, self.element_index, \
            self.xi, self.eta, self.gamma = self.find(x=origin.latitude, y=origin.longitude, z=z, n=1, b_depth=b_depth)

            # MUST subtract 1, the element_index in the point_cloud_file starts from 1.
            self.idx_element = self.element_index - 1
            self.proc_name = get_proc_name(self.idx_processor)
            self._initial_element_frame()
        self._initial_DGFs_N_station()
        # The Lagrange interpolation to get interpolated GF.
        self.interp_Lagrange(self.xi, self.eta, self.gamma)
        if b_verbose:
            print("Station={}, time cost={} s.".format(station.id, time.time() - t0))

        return self.dgf_interp


    def get_greens_function(self, station, origin, b_new_origin=True):
        '''
        Get Greens Function between the station-origin pair.
        Unit: m/N.m

        :param station: An instance of the obspy AttribDict class. For example:
                        station = AttribDict({ 'latitude': 34.0210,
                                                'longitude': -118.287,
                                                'network': 'CI',
                                                'station': 'USC',
                                                'location': '',
                                                'id': 'USC'})

        :param origin:  An instance of the obspy AttribDict class. For example:
                        origin = Origin({'time': '2019-07-04T18:39:44.0000Z',
                                              'latitude': 35.601333,
                                              'longitude': -117.597,
                                              'depth_in_m': 2810.0,
                                              'id': 'evt11056825'})

        :param b_new_origin: True, If acquiring GFs at multiple stations for a same origin to save time.
        '''

        greens = self.get_dgf(station, origin, b_new_origin=b_new_origin)
        return self.get_greens_function_next(greens, station)


    def get_greens_function_next(self, greens, station):
        ''' The next step of get_greens_function()'''
        _, n_force, n_chans = np.shape(greens)
        element_order = ['E', 'N', 'Z']
        st = Stream()

        for i in range(n_force):
            for j in range(n_chans):
                tr = Trace(greens[:, i, j])
                tr.stats.channel = '%s%s' % (element_order[i], element_order[j])
                tr.stats.delta = self.dt
                tr.stats.sampling_rate = int(1.0 / self.dt)
                try:
                    tr.stats.network = station.network
                    tr.stats.station = station.station
                except:
                    pass

                st.append(tr)
        st.id = station.id
        return st


    def get_waveform(self, station, origin, force_enz, b_RTZ=False, b_new_origin=True):
        '''
        Return the synthetic waveform in ENZ (default) or RTZ.
        Unit: m

        :param station: An instance of the obspy AttribDict class. For example:
                        station = AttribDict({ 'latitude': 34.0210,
                                                'longitude': -118.287,
                                                'network': 'CI',
                                                'station': 'USC',
                                                'location': '',
                                                'id': 'USC'})

        :param origin:  An instance of the obspy AttribDict class. For example:
                        origin = Origin({'time': '2019-07-04T18:39:44.0000Z',
                                              'latitude': 35.601333,
                                              'longitude': -117.597,
                                              'depth_in_m': 2810.0,
                                              'id': 'evt11056825'})

        :param force: the force in ENZ [E, N, Z]
        :param b_new_origin: True, If acquiring GFs at multiple stations for a same origin to save time.
        '''

        greens = self.get_dgf(station, origin, b_new_origin=b_new_origin)
        return self.get_waveform_next(greens, station, origin, force_enz, b_RTZ)


    def get_waveform_next(self, greens, station, origin, force_enz, b_RTZ=False):
        ''' The next step of get_waveform()'''

        syn_enz = np.matmul(greens, force_enz).transpose()
        n_chans = 3
        channel_order = ['E', 'N', 'Z']
        channel_name = 'SY'

        # synthetic waveform
        st = Stream()
        for i in range(n_chans):
            tr = Trace(syn_enz[i])
            tr.stats.channel = '%s%s' % (channel_name, channel_order[i])
            tr.stats.delta = self.dt
            tr.stats.sampling_rate = int(1.0 / self.dt)
            st.append(tr)
            try:
                tr.stats.network = station.network
                tr.stats.station = station.station
            except:
                pass

        st.id = station.id

        # Rotation if needed.
        if b_RTZ:
            client = Client()
            res = client.distaz(stalat=station.latitude, stalon=station.longitude,
                            evtlat=origin.latitude, evtlon=origin.longitude)
            back_azimuth = res['backazimuth']
            distance_deg = res['distance']
            distance_m = res['distancemeters']
            st.rotate(method='NE->RT', back_azimuth=back_azimuth)
        return st



# -------------------------------------------------------------------
# Strain Green's Tensor (SGT) database Manager
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import os.path
from seisgen.DPointMgr import DPointCloud
from seisgen.seismic.DSyn import DSyn, RTP_to_DENZ
from seisgen.util_SPECFEM3D import get_proc_name
from seisgen.MTTools.DMomentTensors import DMT_enz
from seisgen.util_SPECFEM3D.ibool_reader import DEnquire_Element
from seisgen.util_SPECFEM3D.xyz_reader import DEnquire_XYZ_GLLs_Element
from seisgen.greens_function.sgt_reader import DEnquire_SGT, read_header_info
from seisgen.math.interp_tools import DCreate_anchors_xi_eta_gamma, DLagrange_interp_sgt, DLagrange_any3D

from obspy.core.util.attribdict import AttribDict
from obspy.core import Stream, Trace
from obspy.clients.iris import Client
from obspy.taup import TauPyModel

import numpy as np
import pandas as pd

import time


class DSGTMgr(DPointCloud):
    '''Strain Green's Tensor (SGT) database Manager'''

    def __init__(self, sgt_database_folder, model3D_folder, point_cloud_file, DLite=False):
        '''
        :param sgt_database_folder:     The directory to the SGT database.
        :param model3D_folder:          The directory to the 3D background model.
        :param point_cloud_file:        The hdf5 file storing the information of user-selected points in the 3D model.
        '''

        self.sgt_database_folder = sgt_database_folder
        self.model3D_folder      = model3D_folder
        self.idx_element         = -1
        self.dt                  = 0
        self.NSPEC               = 0
        # initial parameters of the SGT database.
        if DLite:
            pass
        else:
            self.__initial_paras()
        super().__init__(point_cloud_file)

    def __initial_paras(self):
        '''Function to initial parameters.'''
        hdf5_files = []
        for (dirpath, dirnames, filenames) in os.walk(self.sgt_database_folder):
            for file in filenames:
                if str(file).lower().endswith('.hdf5') or str(file).lower().endswith('.h5'):
                    hdf5_files.append(os.path.join(dirpath, file))

        # Read the Header file to get the information of SGT database.
        for file in hdf5_files:
            try:
                self.dt, self.NSPEC = read_header_info(file)
                break
            except:
                continue

        if 0 == self.dt or 0 == self.NSPEC:
            raise Exception

    def _initial_element_frame(self):
        ''' return the gll information (index, location) at one selected element. '''
        ibool_file = os.path.join(str(self.model3D_folder), str(self.proc_name)+str("_ibool.bin"))
        self.idx_glls = DEnquire_Element(ibool_file,  self.idx_element, self.NSPEC)

        x_glls, y_glls, z_glls = DEnquire_XYZ_GLLs_Element(self.model3D_folder,
                                                           self.idx_processor,
                                                           self.idx_element,
                                                           self.NSPEC)
        self.xyz_glls = np.transpose(np.vstack([x_glls, y_glls, z_glls]))

    def _initial_SGTs_N_station(self):
        '''Return the SGT between origin and station. '''
        dir_string = os.path.join(str(self.sgt_database_folder),
                                     str(self.station.network),
                                     str(self.station.station),
                                     str(self.proc_name))
        sgt_data_path = dir_string + str("_sgt_data.bin")
        sgt_hder_path = dir_string + str("_header.hdf5")
        self.sgts = DEnquire_SGT(sgt_data_path, sgt_hder_path, self.idx_glls)

    def interp_sgt_Lagrange(self, xi, eta, gamma):
        '''Using lagrange method to interpolate the SGT data.'''
        ngll_x = 3
        ngll_y = 3
        ngll_z = 3

        xi_gll, eta_gll, gamma_gll = DCreate_anchors_xi_eta_gamma(ngll_xyz=3)
        h_xi_arr, h_eta_arr, h_gamma_arr = DLagrange_any3D(xi, eta, gamma, xi_gll, eta_gll, gamma_gll)

        self.sgt_interp = DLagrange_interp_sgt(h_xi_arr, h_eta_arr, h_gamma_arr, self.sgts,
                                          ngll_x=ngll_x, ngll_y=ngll_y, ngll_z=ngll_z)

    def set_dt(self, dt):
        '''
        Set time interval.
        Function observed for SeisClient.
        '''
        self.dt = dt
        return self


    def get_sgt(self, station, origin, b_new_origin=True, b_verbose=False):
        '''
        Get the interpolated SGT between the station-origin pair.
        Unit: m/N.m
        :param station: An instance of the obspy AttribDict class. For example:
                        station = AttribDict({ 'latitude': 34.0210,
                                                'longitude': -118.287,
                                                'network': 'CI',
                                                'station': 'USC',
                                                'location': '',
                                                'id': 'USC'})

        :param origin:  An instance of the obspy AttribDict class. For example:
                        origin = Origin({'time': '2019-07-04T18:39:44.0000Z',
                                              'latitude': 35.601333,
                                              'longitude': -117.597,
                                              'depth_in_m': 2810.0,
                                              'id': 'evt11056825'})
        :param b_new_origin: True, If acquiring SGF at multiple stations for a same origin to save time.
        '''

        if b_verbose:
            t0 = time.time()

        self.station = station

        if not self.b_pointcloud_initial:
            raise Exception
        
        try:
            z = origin.depth_in_m
            b_depth = True
        except:
            z = origin.elevation_in_m
            b_depth = False

        if b_new_origin:
            _, _, _, _, \
            _, _, _, \
            self.idx_processor, self.element_index, \
            self.xi, self.eta, self.gamma = self.find(x=origin.latitude, y=origin.longitude, z=z, n=1, b_depth=b_depth)

            # MUST subtract 1, the element_index in the point_cloud_file starts from 1.
            self.idx_element = self.element_index - 1
            self.proc_name = get_proc_name(self.idx_processor)
            self._initial_element_frame()

        self._initial_SGTs_N_station()
        # The Lagrange interpolation to get interpolated SGT.
        self.interp_sgt_Lagrange(self.xi, self.eta, self.gamma)

        if b_verbose:
            print("Station={}, time cost={} s.".format(station.id, time.time()-t0))

        return self.sgt_interp


    def get_greens_function(self, station, origin, b_new_origin=True):
        '''
        Get Greens Function between the station-origin pair.
        Unit: m/N.m

        :param station: An instance of the obspy AttribDict class. For example:
                        station = AttribDict({ 'latitude': 34.0210,
                                                'longitude': -118.287,
                                                'network': 'CI',
                                                'station': 'USC',
                                                'location': '',
                                                'id': 'USC'})

        :param origin:  An instance of the obspy AttribDict class. For example:
                        origin = Origin({'time': '2019-07-04T18:39:44.0000Z',
                                              'latitude': 35.601333,
                                              'longitude': -117.597,
                                              'depth_in_m': 2810.0,
                                              'id': 'evt11056825'})

        :param b_new_origin: Accelerating the extraction of SGT data for multiple stations with the same origin.
        '''

        sgt = self.get_sgt(station, origin, b_new_origin=b_new_origin)
        return self.get_greens_function_next(sgt, station, origin, b_new_origin)


    def get_greens_function_next(self, sgt, station, origin):
        '''The next step of get_greens_function()'''
        client = Client()
        res = client.distaz(stalat=station.latitude, stalon=station.longitude,
                            evtlat=origin.latitude, evtlon=origin.longitude)
        back_azimuth = res['backazimuth']
        azimuth = res['azimuth']
        distance_deg = res['distance']
        distance_m = res['distancemeters']
        stream = self._SGT2GF(sgt, azimuth, back_azimuth)
        stream.id = station.id
        return stream


    def get_waveform(self, station, origin, mt_RTP, b_RTZ=False, b_new_origin=True):
        '''
        Return the synthetic 3-C waveform in RTZ.
        Unit: m

        :param station: An instance of the obspy AttribDict class. For example:
                        station = AttribDict({ 'latitude': 34.0210,
                                                'longitude': -118.287,
                                                'network': 'CI',
                                                'station': 'USC',
                                                'location': '',
                                                'id': 'USC'})

        :param origin:  An instance of the obspy AttribDict class. For example:
                        origin = Origin({'time': '2019-07-04T18:39:44.0000Z',
                                              'latitude': 35.601333,
                                              'longitude': -117.597,
                                              'depth_in_m': 2810.0,
                                              'id': 'evt11056825'})

        :param mt: the moment tensor in RTP [Mrr, Mtt, Mpp, Mrt, Mrp, Mtp]
        :param b_new_origin: Accelerating the extraction of SGT data for multiple stations with the same origin.
        '''

        sgt = self.get_sgt(station, origin, b_new_origin=b_new_origin)
        return self.get_waveform_next(sgt, station, origin, mt_RTP, b_RTZ)


    def get_waveform_next(self, sgt, station, origin, mt_RTP, b_RTZ=False):
        '''The next step of get_waveform()'''

        client = Client()
        res = client.distaz(stalat=station.latitude, stalon=station.longitude,
                            evtlat=origin.latitude, evtlon=origin.longitude)
        back_azimuth = res['backazimuth']
        distance_deg = res['distance']
        distance_m = res['distancemeters']

        # synthetic
        element='SY'
        mt_enz = RTP_to_DENZ(mt_RTP)
        _st = DSyn(mt_enz, sgt, element)
        for _tr in _st:
            _tr.stats.delta = self.dt
            _tr.stats.sampling_rate = int(1.0 / self.dt)

        # waveform in ENZ or RTZ convention
        if b_RTZ:
            _st.rotate(method='NE->RT', back_azimuth=back_azimuth)
        return _st

    def _SGT2GF(self, sgt, azi, ba):
        '''
        Get 3D MT Greens function
        ( Up-South-East convention (USE) by default, which is compatible with MTUQ,
        otherwise North-East-Down convention)
        '''

        mt_rtp = pd.DataFrame({
            "Mrr": np.array([1.0,  0,    0,    0,   0,    0]),
            "Mtt": np.array([0,    1.0,  0,    0,   0,    0]),
            "Mpp": np.array([0,    0,   1.0,   0,   0,    0]),
            "Mrt": np.array([0,    0,   0,     1.0, 0,    0]),
            "Mrp": np.array([0,    0,   0,     0,   1.0,  0]),
            "Mtp": np.array([0,    0,   0,     0,   0,    1.0]),
        })

        stream = Stream()
        mt_orders = ['Mrr', 'Mtt', 'Mpp', 'Mrt', 'Mrp', 'Mtp',]

        for i_mt in mt_orders:
            mt_enz = RTP_to_DENZ(mt_rtp[i_mt].values)
            _st = DSyn(mt_enz, sgt, element='SYN')
            _st.rotate(method='NE->RT', back_azimuth=ba)
            for _tr in _st:
                ch = _tr.stats.channel
                _tr.stats.channel = '%s.%s' % (ch[-1], i_mt)
                _tr.stats._component = ch[-1]
                _tr.stats.delta = self.dt
                _tr.stats.sampling_rate = int(1.0 / self.dt)

            stream += _st
        return stream



# -------------------------------------------------------------------
# Moment tensor
#
# refs
# Tape, W., & Tape, C. (2015). A uniform parametrization of moment tensors.
# Geophysical Journal International, 202(3), 2074–2081. https://doi.org/10.1093/gji/ggv262
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------


import numpy as np

def DCreate_Triple(colatitude, longitude):
    '''
    * Create the normalized triple according to the colat. and long. on the Lune.
    * Equ. (7) in Carl's paper (2015).
    :param colatitude: The colatitude in the Lune in radians (Beta)
    :param longitude:  The longitude in the Lune in radians (Gamma)
    :return: normalized eigenvalues in descent order (a triple on the LUNE) [e1>e2>e3]
    '''

    sin_beta = np.sin(colatitude)
    cos_beta = np.cos(colatitude)
    sin_gamma= np.sin(longitude)
    cos_gamma = np.cos(longitude)

    sqrt3 = 1.73205081
    sqrt2 = 1.41421356
    sqrt6 = 2.44948974

    eigenvalues = np.dot(sqrt6 * np.array([[sqrt3, -1.0, sqrt2],
                                           [0, 2.0, sqrt2],
                                           [-1.0 * sqrt3, -1, sqrt2]]),
                         np.array([sin_beta * cos_gamma, sin_beta * sin_gamma, cos_beta]).transpose())

    return eigenvalues / np.linalg.norm(eigenvalues)


# -------------------------------------------------------------------
# Tools to operate moment tensor and create synthetic waveforms
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import numpy as np
from obspy.core.stream import Stream
from obspy.core.trace import Trace

def RTP_to_DENZ(mt):
    ''' MT from RTP to ENZ   '''
    new_mt = np.zeros_like(mt)
    new_mt[0] = mt[2]
    new_mt[1] = mt[1]
    new_mt[2] = mt[0]
    new_mt[3] = -1.0 * mt[5]
    new_mt[4] = mt[4]
    new_mt[5] = -1.0 * mt[3]
    return new_mt


def DSyn(mt, sgt, element):
    '''
    :param mt:      The moment tensor (MT) in ENZ
    :param SGT:     The strain Greens tensor.
                    The force order: N-E-Z
    :param element: The string of MT component.
    :return:
    '''

    new_mt = mt.copy()
    new_mt[3:] = 2.0 * new_mt[3:]

    n_force = 3
    n_element = 6
    stream = Stream()
    channels = ['N', 'E', 'Z']
    for i in range(n_force):
        trace = Trace(np.dot(sgt[:, i, :], new_mt))
        trace.stats.channel = str(element)+str(channels[i])
        stream.append(trace)
    return stream






# -------------------------------------------------------------------
# Calculate the orientation matrix for the given focal mechanism.
#
# refs
# Tape, W., & Tape, C. (2015). A uniform parametrization of moment tensors.
# Geophysical Journal International, 202(3), 2074–2081. https://doi.org/10.1093/gji/ggv262
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------


import numpy as np

def DRotx(angle):
    ''' Get rotation matrix at the x axis. angle in Radians. '''
    return np.array([[1, 0, 0],
                     [0, np.cos(angle), -1 * np.sin(angle)],
                     [0, np.sin(angle), np.cos(angle)]])

def DRoty(angle):
    ''' Get rotation matrix at the y axis. angle in Radians. '''
    return np.array([[np.cos(angle), 0, np.sin(angle)],
                     [0, 1, 0],
                     [-1 * np.sin(angle), 0, np.cos(angle)]])


def DRotz(angle):
    ''' Get rotation matrix at the z axis. angle in Radians. '''
    return np.array([[np.cos(angle), -1 * np.sin(angle), 0],
                     [np.sin(angle), np.cos(angle), 0],
                     [0, 0, 1]])


def DRotationMatrix(strike, dip, rake):
    '''
    * Calculate the rotation matrix for the given focal mechanism.
    * Using the Equ (10) in Carl's Paper (2015).
    :param strike: The strike in radians. k[kappa] ranging [0, 2pi]
    :param dip:     The dip in radians. the theta ranging [0, pi/2]
    :param rake:    The slip angle or the rake angle. the sigma ranging [-pi/2, pi/2]
    :return: the rotation matrix U(k(=strike), sigma(=rake), theta(=dip))
    '''
    Z_k = DRotz(-1.0 * strike)
    X_theta = DRotx(dip)
    Z_delta = DRotz(rake)
    tmp = np.dot(Z_k, X_theta)
    V = np.dot(tmp, Z_delta)

    Y_pi4 = DRoty(np.pi/-4.0)
    return np.dot(V, Y_pi4).round(6)


# -------------------------------------------------------------------
# XYZ reader.
# *x.bin, *y.bin, *z.bin
#
# Author: Liang Ding
# Email: myliang.ding@mail.utoronto.ca
# -------------------------------------------------------------------

import os.path
from seisgen.util_SPECFEM3D import get_proc_name, CONSTANT_INDEX_27_GLL
from seisgen.util_SPECFEM3D.ibool_reader import read_ibool_by_scipy
from scipy.io import FortranFile


def read_xyz_bin_by_scipy(file_path):
    '''
    * read XYZ bin file by using the SciPy package.

    :param file_path: The path of the XYZ bin file.
    :param file_type: The data type of the XYZ bin.

    * x,y,z -- float32

    :return: The whole data.
    '''

    data_type = 'float32'

    try:
        f = FortranFile(file_path, 'r')
        dat = f.read_reals(dtype=data_type)
        f.close()
    except:
        print("Unable to open file: ", str(file_path))
        return None
    return dat


def DEnquire_XYZ_GLLs_Element(data_dir, idx_processor, idx_element, NSPEC):
    '''
    return the x, y, z of the 27 GLL points where the SGT been stored in the selected element.

    :param data_dir:        The dir of the *.bin files.
    :param idx_proc:        The index of the processor. INT
    :param idx_element:     The index of the element in the processor. INT
    :param NSPEC_PER_SLICE:    The number of the element in the processor. INT
    :return:                The x, y, and z array of the GLL points.
    '''

    proc_name = get_proc_name(idx_processor)
    ibool_file = os.path.join(str(data_dir),  str(proc_name) + "_ibool.bin")
    x_file = os.path.join(str(data_dir),  str(proc_name) + "_x.bin")
    y_file = os.path.join(str(data_dir), str(proc_name) + "_y.bin")
    z_file = os.path.join(str(data_dir), str(proc_name) + "_z.bin")

    ibool = read_ibool_by_scipy(ibool_file, NSPEC)
    
    x = read_xyz_bin_by_scipy(x_file)
    y = read_xyz_bin_by_scipy(y_file)
    z = read_xyz_bin_by_scipy(z_file)
    glls_idx = ibool[idx_element][CONSTANT_INDEX_27_GLL]

    return x[glls_idx], y[glls_idx], z[glls_idx]




import setuptools

setuptools.setup(
    name="seisgen",
    version="0.2.1",
    author="Liang Ding",
    author_email="myliang.ding@mail.utoronto.ca",
    description="The theoretical seismic waveform generation tool",
    long_description="The theoretical seismic waveform generation tool based on the stored Strain Green's Tensor (SGT)"
                     " database that is built by using the 3D background model and SPECFEM3D package.",
    long_description_content_type="text/markdown",
    url="https://github.com/Liang-Ding/seisgen",
    project_urls={
        "Bug Tracker": "https://github.com/Liang-Ding/seisgen/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords=[
        "seismology"
    ],
    # package_dir={"": "seisgen"},
    python_requires='>=3.6.0',
    install_requires=[
        "numpy", "scipy", "obspy",
        "h5py",
    ],
    # packages=setuptools.find_packages(where="seisgen"),
    packages=setuptools.find_packages(),
)
